# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
import os

from datetime import date
from django.utils.translation import ugettext as _, ugettext_lazy
from basesoft.web.contrib.file.controllers import ImportFileController
from basesoft.web.core.importers import Utf8CsvImporter, XlsxImporter
from sqlalchemy.sql.expression import and_
from we.core.importers import WeImportException
from we.declaration.controllers import WeDeclarationController
from we.declaration.models import WeDeclaration
from we.member.models import WeLpp, WeHqppLicense
from we.product.models import WeHqppRegulationPeriodProduct, WeHqppProduct
from we.tbs.models import WeTbsRegulationPeriod


class WeDeclarationAbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.declaration.Declaration'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('Declaration')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = WeDeclarationCsvImporter

        elif file_extension == '.xlsx':
            importer = WeDeclarationXlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_human_readable_settings(self, import_settings):
        rv = {}

        dc = WeDeclarationController(self.sa_session, account_id=self.get_account().id)
        hqpp = dc.get_member_licenses(import_settings['hqpp_id']).pop()

        rv[_('HQPP')] = hqpp.name
        return rv


class WeDeclarationImporter(WeDeclarationAbstractImporter):

    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = 7
    STOP_AFTER_NUMBER_EMPTY_ROWS = 3

    import_state = None
    callback = None

    def __init__(self, session, **kwargs):
        ImportFileController.__init__(self, session, **kwargs)
        self.import_state = {}
        self.row_number = 1

    @staticmethod
    def get_notification_callback():
        return {
            'js_module': 'we_declaration_manager'
        }

    @staticmethod
    def get_failure_notification_callback():
        return {
            'js_module': 'we_declaration_import_wizard'
        }

    @staticmethod
    def get_notification_subject():
        return None

    @staticmethod
    def get_notification_text():
        return None

    @staticmethod
    def get_notification_update_notify():
        return ['we.product.FileImportCustomQuery.newData', 'we.declaration.WeDeclarationCustomQuery.newData']

    def _import_row(self, row, controller, dryrun=False, hqpp_id=None, row_count=None, **kwargs):

        if row[6] in [None, '']:
            return
        try:
            lpp_id = int(row[1])
            year = int(row[2])
            month = int(row[3])
            sku = unicode(row[4])
            is_b2b = row[5] if row[5] not in [None, ''] else None
            quantity = int(row[6])
        except ValueError:
            raise WeImportException(_('Could not convert data to an integer.'))

        hqpp = self.sa_session.query(WeHqppLicense).filter(WeHqppLicense.license_id == hqpp_id).first()
        if hqpp is None:
            raise WeImportException(_('There is no HQPP with ID %s.' % hqpp_id))

        lpp = self.sa_session.query(WeLpp).filter(WeLpp.id == lpp_id, WeLpp.hqpp_id == hqpp_id).first()
        if lpp is None:
            raise WeImportException(_("Producer %(producer)s has no access to LPP with ID %(id)s " % {'producer': hqpp.name, 'id': lpp_id}))

        if is_b2b in [None, '']:
            is_b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE
        elif is_b2b.lower() == 'b2b':
            is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_TRUE
        else:
            is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_FALSE

        if is_b2b == WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
            if lpp.tbs.get_config('we.tbs.require.end.user.distinction') is True:
                raise WeImportException(_("Line %(line)s with SKU %(sku)s "
                                          "requires B2B/B2C distinction." % {'line': self.row_number, 'sku': sku}))

        if not self.import_state.has_key(lpp_id):
            self.import_state[lpp_id] = {}
        if not self.import_state[lpp_id].has_key(year):
            self.import_state[lpp_id][year] = {}
        if not self.import_state[lpp_id][year].has_key(month):
            self.import_state[lpp_id][year][month] = {}

        # add line to the state
        self.import_state[lpp_id][year][month]['lpp_id'] = lpp_id
        self.import_state[lpp_id][year][month]['year'] = year
        self.import_state[lpp_id][year][month]['month'] = month
        if not self.import_state[lpp_id][year][month].has_key('lines'):
            self.import_state[lpp_id][year][month]['lines'] = []

        line = {
            'line_id': None,
            'sku': sku,
            'quantity': quantity,
            'is_b2b': is_b2b
        }
        self.import_state[lpp_id][year][month]['lines'].append(line)

        return True

    def _post_read_rows(self, controller, **kwargs):
        """
        called after above all _import_row() is done
        """

        for lpp_id in self.import_state.keys():
            for year in self.import_state[lpp_id].keys():
                for month in self.import_state[lpp_id][year].keys():
                    loaded_state = {}  # reset loaded state, as it is a by Declaration month state (#500)
                    _id = None

                    lpp = self.sa_session.query(WeLpp).filter(WeLpp.id == lpp_id).first()

                    declaration = self.sa_session.query(WeDeclaration)\
                        .filter(and_(WeDeclaration.lpp == lpp,
                                     WeDeclaration.year == year,
                                     WeDeclaration.month == month)).first()

                    if declaration is not None:
                        loaded_state = controller.load_declaration(declaration.id)
                        _id = declaration.id

                    # check for regulation period
                    _filter = and_(WeTbsRegulationPeriod.tbs == lpp.tbs,
                                   WeTbsRegulationPeriod.start_date <= date(year, month, 1),
                                   WeTbsRegulationPeriod.archived == 0)
                    regulation_period = self.sa_session.query(WeTbsRegulationPeriod).filter(_filter)\
                        .order_by(WeTbsRegulationPeriod.start_date.desc()).first()

                    if regulation_period is None:
                        raise WeImportException(_("Cannot import declaration. There are no regulation "
                                                  "periods for take-back scheme %s." % lpp.tbs.name))

                    # create a state from old lines and new lines
                    lines_state = self._create_declaration_state(lpp.hqpp, regulation_period,
                                                                 self.import_state[lpp_id][year][month]['lines'],
                                                                 loaded_state)

                    self.import_state[lpp_id][year][month]['lines'] = lines_state
                    controller.save_or_update_declaration(_id, self.import_state[lpp_id][year][month])

    def _create_declaration_state(self, hqpp, regulation_period, lines, loaded_state):

        for line in lines:
            matched = False

            # if sku is in loaded state lines set the line id and the hqpp period product
            # else add new line
            if loaded_state.has_key('lines'):
                for l in loaded_state['lines']:
                    if line['sku'].lower() == l['sku'].lower():

                        if line['is_b2b'] != l['is_b2b']:
                            continue

                        line['hqpp_period_product_id'] = l['hqpp_period_product_id']
                        line['line_id'] = l['line_id']
                        matched = True

            if matched is False:
                # get the hqpp product corresponding to the SKU
                product = self.sa_session.query(WeHqppProduct).filter(
                    WeHqppProduct.hqpp == hqpp,
                    WeHqppProduct.sku == line['sku']
                ).one()

                product_tbs_category = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
                    and_(
                        WeHqppRegulationPeriodProduct.product == product,
                        WeHqppRegulationPeriodProduct.tbs_regulation_period == regulation_period,
                        WeHqppRegulationPeriodProduct.is_b2b == line['is_b2b']
                    )
                ).first()
                line['hqpp_period_product_id'] = product_tbs_category.id
        return lines

class WeDeclarationCsvImporter(WeDeclarationImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        return self._read_row(file_path)


class WeDeclarationXlsxImporter(WeDeclarationImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):

        active_sheet = self._read_row(file_path)

        for row in active_sheet.iter_rows():
            values = []
            cell_i = 0
            for cell in row:
                cell_i += 1
                if cell_i > 7:
                    break

                value = cell.internal_value
                if type(value) == float:
                    value = int(value)
                values.append(value)

            yield values

import_file_map = {
    'we.declaration.Declaration': WeDeclarationAbstractImporter,
}