# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
import os

from django.conf import settings
from django.http import HttpResponse
from django.utils.translation import ugettext as _

from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins

from we.declaration.controllers import WeDeclarationController
from we.declaration.importers import WeDeclarationXlsxImporter, WeDeclarationCsvImporter
from we.member.models import WeHqppLicense


class WeDeclarationRpcHandler:

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self.wdc = WeDeclarationController(self.sa_session, account_id=self.account_id, request=request)

    def save_or_update_declaration(self, declaration_id, state):
        self.wdc.save_or_update_declaration(declaration_id, state)

    def load_declaration(self, declaration_id):
        return self.wdc.load_declaration(declaration_id)

    def get_declaration_id_for_period(self, lpp_id, year, month):
        return self.wdc.get_declaration_id_for_period(lpp_id, year, month)

    def get_regulation_period(self, lpp_id, year, month):
        rp = self.wdc.get_regulation_period(lpp_id, int(year), int(month))
        if rp is None:
            return {}
        return {
            'id': rp.id,
            'title': rp.title
        }

    def import_declaration_file(self, file_hash, hqpp_id):

        self.wdc.assert_permissions('we_manage_declaration')

        if hqpp_id in [None, '']:
            _licenses = self.wdc.get_member_licenses()
            if len(_licenses) > 1:
                raise PublicUserException(_("Please, select a producer."))
            _licenses = _licenses.pop()
            assert isinstance(_licenses, WeHqppLicense)
            hqpp_id = _licenses.id

        tmp_file = file_hash[0]
        base_filename, file_extension = self.wdc.get_file_extension(tmp_file.file_name, tmp_file.file_hash)

        if file_extension in ('.csv', '.txt'):
            importer = WeDeclarationCsvImporter

        elif file_extension == '.xlsx':
            importer = WeDeclarationXlsxImporter

        else:
            raise PublicUserException(_('Cannot import for extension: %s' % file_extension))

        file_path = os.path.join(settings.TMP_FILE_REPOSITORY, tmp_file.file_hash)
        importer = importer(self.sa_session, account_id=self.account_id)

        # validate SKUs
        missing_skus = self.wdc.validate_declaration(importer, file_path, hqpp_id)
        if len(missing_skus) > 0:
            return missing_skus

        return self.wdc.import_declaration(tmp_file, importer, file_path, hqpp_id=hqpp_id)

    def get_available_months(self, lpp_id, year):
        month = self.wdc.today().month - 1
        if month == 0:
            return []

        y = self.wdc.today().year
        rv = []
        months = self.wdc.get_months_map()

        periods = self.wdc.get_saved_declaration_periods(lpp_id, year)
        if int(year) < y:
            month = 12
        for i in xrange(1, month + 1):

            if i in periods:
                continue
            rv.append({
                'value': i,
                'label': months[i - 1]
            })

        return rv

    def get_all_months(self):
        return self.wdc.get_months()

    def get_years(self):
        year = self.wdc.today().year
        rv = []

        for i in xrange(year - 2, year + 1):
            rv.append({
                'value': i,
                'label': i
            })

        return rv

    def validate_declaration_form(self, args):
        form_handler = FormValidator({
            'lpp_id': [(plugins.Required, {})],
            'sku': [(plugins.Required, {})],
            'quantity': [(plugins.Required, {})],
            'year': [(plugins.Required, {})],
            'month': [(plugins.Required, {})],
        })

        return form_handler.validate(args)

    def validate_declaration_export_form(self, args):
        return FormValidator({
            'hqpp_id': [(plugins.Required, {})],
            'start_year': [(plugins.Required, {})],
            'start_month': [(plugins.Required, {})],
        }).validate(args)


def export_declaration(request, hqpp_id, year, month):
        """
        Writes data to excel file. User for Customdatatable data export.
        """
        sa_session = request.Session()

        wpc = WeDeclarationController(sa_session, account_id=request.session['user_id'], request=request)
        f_name = wpc.export_declaration(int(hqpp_id), int(year), int(month))

        response = HttpResponse(open(f_name, 'r'), mimetype='application/excel')
        response['Content-Disposition'] = 'attachment; filename=%s_.xlsx' % f_name

        return response

handlers = (WeDeclarationRpcHandler, )