# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy import Integer, Column, DateTime, ForeignKey, UniqueConstraint, SmallInteger, Numeric
from sqlalchemy.orm import mapper, relation

from basesoft.web.contrib.license import models as license_models
from basesoft.sqlalchemy.table import IQTable

from we.declaration import models as declaration_models

__mapping_defined = False
__tables_defined = False
tables = {}


def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False


def get_tables():
    return tables


def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    tblWeDeclaration = IQTable('tblWeDeclaration', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('lpp_id', Integer, ForeignKey('tblWeLpp.id', ondelete='CASCADE'), nullable=False),
        Column('currency_id', ForeignKey('tblCurrency.id', ondelete='CASCADE')),
        Column('declaration_order_id', ForeignKey('tblWeDeclarationOrder.order_id', ondelete='CASCADE'), nullable=True),

        Column('year', SmallInteger, nullable=False),
        Column('month', Integer),
        Column('total_weight', Numeric(20, 5), nullable=False),
        Column('total_fee', Numeric(20, 5), nullable=False),
        Column('total_quantity', Integer, nullable=False),

        Column('status', SmallInteger, nullable=False, default=declaration_models.WeDeclaration.STATUS_COMPLETE),

        Column('export_date', DateTime()),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('lpp_id', 'year', 'month')
    )

    tblWeDeclarationLine = IQTable('tblWeDeclarationLine', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('position', Integer),
        Column('declaration_id', Integer, ForeignKey('tblWeDeclaration.id', ondelete='CASCADE'), nullable=False),
        Column('hqpp_period_product_id', Integer, ForeignKey('tblWeHqppRegulationPeriodProduct.id', ondelete='RESTRICT'), nullable=False),

        Column('quantity', Integer, nullable=False),
        Column('product_total_weight', Numeric(20, 5)),
        Column('product_total_fee', Numeric(20, 5)),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('id', 'declaration_id'),
        UniqueConstraint('hqpp_period_product_id', 'declaration_id')
    )

    tables['tblWeDeclaration'] = tblWeDeclaration
    tables['tblWeDeclarationLine'] = tblWeDeclarationLine

    __tables_defined = True


def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from we.member import models as member_models
    from we.product import models as product_models
    from we.order import models as we_order_models
    from basesoft.web.contrib.order import models as order_models
    from basesoft.web.contrib.notification import models as notification_models

    mapper(declaration_models.WeDeclaration, tables['tblWeDeclaration'],
        properties={
            'lpp': relation(member_models.WeLpp),
            'currency': relation(order_models.Currency),
            'lines': relation(declaration_models.WeDeclarationLine, backref='declaration'),
            'declaration_order': relation(we_order_models.WeDeclarationOrder, backref="declarations"),
        }
    )

    mapper(declaration_models.WeDeclarationLine, tables['tblWeDeclarationLine'],
        properties = {
            'hqpp_period_product': relation(product_models.WeHqppRegulationPeriodProduct, order_by=tables['tblWeDeclarationLine'].c.position)
        }
    )

    mapper(declaration_models.WeDeclarationReminderNotification,
           inherits=notification_models.Notification,
           polymorphic_identity='WeDeclarationReminderNotification'
    )

    mapper(declaration_models.WeDeclarationResubmitNotification,
           inherits=notification_models.Notification,
           polymorphic_identity='WeDeclarationResubmitNotification'
    )

    __mapping_defined = True


def fill_data(conn):
    pass


def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'WEEE Europe declaration')


def app_permissions():
    from django.utils.translation import ugettext as _
    return [

        license_models.Permission(u'we_manage_declaration', u"""
            This gives the right to manage Producer partner declarations
            """, _("Manage declarations")),

        license_models.Permission(u'we_view_declaration', u"""
            This gives the right to view declarations
            """, _("View declarations")),
    ]

