# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy.sql.expression import func, and_

from django.utils.translation import ugettext as _

from basesoft.web.contrib.contact.models import CompanyAddress, Address
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.order.models import Currency
from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.contrib.account.controllers import AccountController
from basesoft.web.contrib.account.models import AccountActivation
from basesoft.web.core.util.controllers import transaction_decorator
from basesoft.web.contrib.license.controllers import LicenseController

from we.core.controllers import AbstractWeController
from we.member.models import WeHqppLicense, WeLpp, WeCompany, WeContact
from we.product.controllers import WeProductController
from we.tbs.models import WeTbsLicense


class WeMemberController(LicenseController, AbstractWeController):

    _instance_class = WeContact

    def get_member(self, _id):
        self.assert_permissions('we_manage_member_license')
        return self.get_member_licenses(member_id=_id).pop()

    def get_members(self):
        self.assert_permissions('we_manage_member_license')
        options = []

        for m in self.get_member_licenses():
            options.append({
                'value': m.id,
                'label': m.name
            })

        return options

    def load_member(self, _id):
        assert isinstance(_id, (int, long))
        self.assert_permissions('we_manage_member_license')

        state = self.load_master_data_state(_id)
        state['company_js_module'] = 'we_company_form'
        state['address_js_module'] = 'we_address_form'
        state['person_js_module'] = 'we_person_form'
        state['bank_account_js_module'] = 'we_bank_account_form'
        return state

    def save_or_update_member(self, _id, state):
        if _id is None:
            return self.save_member(state)
        state['dbID'] = _id
        auto_approve = self.has_permissions('license_masterdata_change_approval')
        if self.has_permissions('we_create_member_license'):
            return self._update_master_data(self.get_member(_id), state, auto_approve=auto_approve)
        return self.update_master_data(_id, state, auto_approve=auto_approve)

    @transaction_decorator
    def save_member(self, state):
        assert isinstance(state, dict)
        self.assert_permissions('we_create_member_license')

        self.logger.info('save_member: %s' % state['company']['name'])

        _license = self._create_new_license(name=state['company']['name'], polymorphic_type=WeHqppLicense, role_name='WE Producer', add_account=False, link_role=True)

        if state['bank_data'].has_key('company_bank_account_address'):
            state['bank_data']['address_data'] = state['bank_data']['company_bank_account_address']
        self.logger.info('calling update_master_data')
        _license = self._update_master_data(_license, state, auto_approve=self.has_permissions('license_masterdata_change_approval'))
        return _license

    @transaction_decorator
    def add_lpp_to_hqpp(self, hqpp_id, state, company_id=None):
        """
        A HQPP can have multiple local producer partners at the same TBS. We verify that the VAT NO is not already used.
        """
        self.assert_permissions('we_manage_lpp_companies')
        hqpp = self.get_member(hqpp_id)

        if len(state['lpp']['tbses']) == 0:
            raise PublicUserException(_("At least one take-back scheme must be selected."))

        for lpp in hqpp.lpps:
            if lpp.company.vat_no == state['vat_no'] and company_id is None:
                raise PublicUserException(_('The VAT number supplied is already part of this HQPP'))

        company = self._save_or_update_company(id=company_id, company_data=state, instance_class=WeCompany)

        lpp_ids = [int(row['lpp_id']) for row in state['lpp']['tbses'] if row['lpp_id'] not in ['', None]]

        # disable LPP => TBS link if lpp_id not in state
        if len(lpp_ids) > 0:
            _where = """ id NOT IN :lpp_ids;"""
            if len(lpp_ids) == 1:
                lpp_ids = lpp_ids[0]
                _where = """ id != :lpp_ids;"""
            self.sa_session.execute("""
                                UPDATE tblWeLpp SET status = :status WHERE company_id = :company_id AND
                                """ + _where, {
                                    'status': WeLpp.STATUS_INACTIVE,
                                    'lpp_ids': lpp_ids,
                                    'company_id': company_id
                                })

        for row in state['lpp']['tbses']:
            if row['lpp_id'] in [None, '']:
                lpp = WeLpp()
                tbs = self.get_tbs(int(row['tbs_id']))
                lpp.tbs = tbs
                lpp.status = WeLpp.STATUS_NEW
            else:
                lpp = self.get_local_producer_partner(row['lpp_id'])

            lpp_relation_id = row['lpp_relation_id']
            if lpp_relation_id is None or lpp_relation_id.strip() == '':
                lpp_relation_id = None

            invoice_relation_id = row['invoice_relation_id']
            if invoice_relation_id is None or invoice_relation_id.strip() == '':
                invoice_relation_id = None

            lpp.hqpp = hqpp
            lpp.name = state['name']
            lpp.company = company
            lpp.lpp_relation_id = lpp_relation_id
            lpp.invoice_relation_id = invoice_relation_id

            if row['use_as_ar'] is True:
                lpp.authorized_representative = company
            elif lpp.authorized_representative == company:
                lpp.authorized_representative = None

            self.sa_session.add(lpp)

        return lpp

    @transaction_decorator
    def register_member(self, master_data, account_data=None):
        self.assert_permissions('we_create_member_license')
        assert isinstance(master_data, dict)

        # create a new license.
        member = self.save_or_update_member(None, master_data)

        # no account data, skip this step.
        if account_data is None:
            return member

        assert isinstance(account_data, dict)

        person = member.company.contact_person
        if not account_data['person']['is_same_person_as']:
            person = self._save_or_update_person(None, account_data['person'])

        ac = AccountController(self.sa_session, account_id=self.get_account().id, request=self.request)
        password = ac.generate_random_password(12) # set a random password.
        send_link = True
        if not account_data['password']['send_link']:
            send_link = False
            password = account_data['password']['new_password']
            verify = account_data['password']['new_password_verify']
            if password != verify:
                raise PublicUserException(_('Verification password does not match supplied new password'))

        # create a new account for the member
        account = ac.create_account(account_data['account_details']['account_name'], password=password, person=person)
        ac.save_or_update_group_account(member.default_group, account, is_default=True, is_active=True)
        if send_link:
            ac.save_account_activation(account, activation_type=AccountActivation.ACTIVATION_TYPE_NEW_ACCOUNT)

        return member

    def load_company_details(self, company_id):
        self.assert_permissions('we_manage_lpp_companies')
        lpps = self.get_lpp_with_all_tbses(company_id)
        company_state = self.load_company_state(lpps[0].company if len(lpps) > 0 else company_id)

        state = {
            'company': company_state,
            'member': {
                'name': lpps[0].hqpp.name if len(lpps) > 0 else ''
            }
        }

        state['lpp'] = {
            'tbses': []
        }

        for lp in lpps:
            state['lpp']['tbses'].append({
                'tbs_id': lp.tbs_id,
                'lpp_relation_id': lp.lpp_relation_id,
                'invoice_relation_id': lp.invoice_relation_id,
                'lpp_id': lp.id,
                'use_as_ar': True if lp.company == lp.authorized_representative else False
            })

        if company_state.has_key('address_data'):
            state['address_data'] = company_state['address_data']

        if company_state.has_key('bank_data'):
            state['bank_data'] = company_state['bank_data']

        if company_state.has_key('contact_person_data'):
            state['contact_person_data'] = company_state['contact_person_data']

        return state

    def load_lpps(self):
        """
        Get all LPPs that the current user has access to.
        :return: List of dictionaries for Select element.
        """
        self.assert_permissions('we_view_lpp_companies')

        lpps = self.get_local_producer_partners(status=WeLpp.STATUS_ACTIVE)
        options = []

        for lpp in lpps:
            options.append({
                'label': lpp.name + ' - ' + lpp.tbs.name,
                'value': lpp.id
            })

        return options

    def save_or_update_company_details(self, company_id, state):
        self.assert_permissions('we_manage_lpp_companies')

        hqpp_id = state['lpp']['member_id'] if state.has_key('lpp') else None

        if company_id is None and hqpp_id is not None:
            return self.add_lpp_to_hqpp(hqpp_id, state)

        if company_id is not None and hqpp_id is not None:
            return self.add_lpp_to_hqpp(hqpp_id, state, company_id=company_id)

        if company_id is None and hqpp_id is None:
            return self._save_or_update_company(id=None, company_data=state, instance_class=WeCompany)

        # update company info
        return self._save_or_update_company(id=company_id, company_data=state, instance_class=WeCompany)

    @transaction_decorator
    def save_companies_linked_to_tbs(self, hqpp_id, state):
        self.assert_permissions('we_manage_lpp_companies')

        # get member
        hqpp = self.get_member_licenses(hqpp_id).pop()

        for row in state['tbs_links']:
            if row['lpp_id'] in [None, '']:
                lpp = WeLpp()
            else:
                lpp = self.sa_session.query(WeLpp).filter(WeLpp.id == int(row['lpp_id'])).one()

            company_id = int(row['lpp_company_id'])

            # check if we have LPP for that company_id... we should not have more than one record without TBS
            # this is only when the company is newly added
            lp = self.sa_session.query(WeLpp).filter(and_(WeLpp.company_id == company_id, WeLpp.tbs == None)).first()

            # if that LPP is not linked to a TBS yet, do it first
            if lp is not None:
                lpp = lp

            if row['ar_company_id'] not in [None, '']:
                lpp.authorized_representative = self.sa_session.query(WeCompany).filter(WeCompany.id == int(row['ar_company_id'])).one()

            if row['invoice_company_id']not in [None, '']:
                lpp.invoice_company = self.sa_session.query(WeCompany).filter(WeCompany.id == int(row['invoice_company_id'])).one()

            lpp.tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.license_id == int(row['tbs_id'])).one()

            lpp.company = self.sa_session.query(WeCompany).filter(WeCompany.id == company_id).one()

            if lpp.hqpp is None:
                lpp.hqpp = hqpp
            lpp.name = lpp.company.name

            self.sa_session.add(lpp)

        return True

    def get_companies_per_hqpp_linked_to_tbs(self, hqpp_id):
        self.assert_permissions('we_manage_lpp_companies')
        companies = self.sa_session.query(WeLpp).filter(WeLpp.hqpp_id == hqpp_id).all()

        rv = {}
        rv['tbs_links'] = []
        for c in companies:
            rv['tbs_links'].append({
                'lpp_id': c.id,
                'tbs_id': c.tbs.id,
                'tbs_name': c.tbs.company.name,
                'lpp_company_id': c.company.id,
                'ar_company_id': None if c.ar_company_id is None else c.authorized_representative.id,
                'invoice_company_id': None if c.invoice_company_id is None else c.invoice_company.id,
            })

        return rv

    def delete_lpp(self, lpp_id):
        self.assert_permissions('we_manage_lpp_companies')
        lpp = self.get_local_producer_partner(lpp_id)

        if lpp.status not in [WeLpp.STATUS_NEW, WeLpp.STATUS_PENDING]:
            raise PublicUserException(_('Cannot delete a previously enabled local producer partner.'))

        self.logger.info('Deleted local producer partner %s' % lpp)
        self.sa_session.delete(lpp)
        self.sa_session.flush()

    @transaction_decorator
    def save_lpp_status(self, lpp_id, status):
        self.assert_permissions('we_manage_lpp_companies')
        if status in [None, '']:
            return False
        lpp = self.get_local_producer_partner(lpp_id)

        self.logger.info('LPP status changed to %s' % status)
        lpp.status = status
        self.sa_session.add(lpp)
        return True

    def get_tbs_and_currency(self, lpp_id):
        lpp = self.get_local_producer_partner(lpp_id)
        currency_id = self.sa_session.query(func.iqGetCurrencyIdForCountry(lpp.tbs.company.addresses[0].address.country.id, self.today())).one()[0]

        currency = self.sa_session.query(Currency).filter(Currency.id == currency_id).first()

        if currency is None:
            raise PublicUserException('There is no currency for country %s' % lpp.tbs.company.addresses[0].address.country.name)

        symbol = currency.symbol
        if not symbol:
            symbol = currency.code

        return {
            'lpp_name': lpp.name,
            'tbs_name': lpp.tbs.name,
            'currency_symbol': symbol
        }

    def get_all_companies(self, contains, value, limit, hqpp_id, tbs_id):

        tbs = self.get_tbs_licenses(int(tbs_id)).pop()

        _where = WeCompany.is_temp == False

         # if value is provided get the company with that ID
        if value not in [None, '']:
            return [self.get_company(int(value))]

        if contains not in [None, '']:
            _where = and_(_where, WeCompany.name.like('%%%s%%' % contains))
            return self.sa_session.query(WeCompany).filter(_where).limit(limit).all()

        _where = and_(_where, WeHqppLicense.license_id == None, WeTbsLicense.license_id == None, WeLpp.id == None)
        companies = self.__query_for_companies(_where, limit)

        # if we don't want all LPPs filter them by country
        lpps = self.sa_session.query(WeLpp)\
            .join(WeHqppLicense, and_(WeLpp.hqpp_id == WeHqppLicense.license_id, WeHqppLicense.name != WeLpp.name))\
            .filter(WeLpp.hqpp_id == hqpp_id).limit(limit).all()
        # get the HQPP and if the country is the same add it
        member = self.get_member_licenses(hqpp_id).pop()
        companies.append(member.company)

        # if we want companies which are not LPP add TBS as well
        companies.append(tbs.company)

        for lpp in lpps:
            companies.append(lpp.company)

        return companies

    def get_ar_companies(self, contains, value, limit, hqpp_id, tbs_id):

        if tbs_id is None:
            return []

        tbs = self.get_tbs_licenses(int(tbs_id)).pop()
        country = tbs.company.addresses[0].address.country

        _where = WeCompany.is_temp == False

         # if value is provided get the company with that ID
        if value not in [None, '']:
            return [self.get_company(int(value))]

        if contains not in [None, '']:
            _where = and_(_where, WeCompany.name.like('%%%s%%' % contains))
            return self.__query_for_companies(_where, limit, country.id)

        # get third party companies
        _where = and_(_where, WeHqppLicense.license_id == None, WeTbsLicense.license_id == None, WeLpp.id == None)
        companies = self.__query_for_companies(_where, limit, country.id)

        # if we don't want all LPPs filter them by country
        lpps = self.sa_session.query(WeLpp)\
            .join(WeHqppLicense, and_(WeLpp.hqpp_id == WeHqppLicense.license_id, WeHqppLicense.name != WeLpp.name))\
            .join(WeCompany, WeLpp.company_id == WeCompany.id).join(CompanyAddress)\
            .join(Address, and_(CompanyAddress.address_id == Address.id, Address.country_id == country.id))\
            .filter(and_(Country.id == country.id, WeLpp.hqpp_id == hqpp_id)).all()

        # HQPP can be AR as well but it needs to be in the TBS country
        member = self.get_member_licenses(hqpp_id).pop()
        if member.company.addresses[0].address.country == country:
            companies.append(member.company)

        companies.append(tbs.company)

        for lpp in lpps:
            companies.append(lpp.company)

        return companies

    def get_companies_acting_as_lpp(self, contains, value, limit, hqpp_id, exclude):

        _where = and_(WeCompany.is_temp == False, WeLpp.hqpp_id == hqpp_id)

         # if value is provided get the company with that ID
        if value not in [None, '']:
            return [self.get_company(int(value))]

        companies = []
        member = self.get_member_licenses(hqpp_id).pop()

        if contains not in [None, '']:
            _where = and_(_where, WeCompany.name.like('%%%s%%' % contains))

        if exclude is not None:
            _where = and_(_where, ~WeCompany.id.in_(exclude))
            if member.company.id not in exclude:
                companies.append(member.company)
        else:
            companies.append(member.company)

        lpps = self.sa_session.query(WeCompany).join(WeLpp, WeCompany.id == WeLpp.company_id)\
            .join(WeHqppLicense, and_(WeHqppLicense.license_id == WeLpp.hqpp_id, WeHqppLicense.name != WeLpp.name))\
            .filter(_where).limit(limit).all()

        return companies + lpps

    def get_all_tbs(self, hqpp_id):
        lpps = self.get_local_producer_partners(hqpp_id)
        tbses = []
        options = []
        for lpp in lpps:
            if lpp.tbs not in tbses:
                options.append({
                    'label': lpp.tbs.name,
                    'value': lpp.tbs.id
                })
                tbses.append(lpp.tbs)
        return options

    def get_tbses_for_product_import(self, hqpp_id):
        lpps = self.get_local_producer_partners(hqpp_id)
        tbses = []
        options = []

        pc = WeProductController(self.sa_session, account_id=self.account_id)
        for lpp in lpps:
            if lpp.tbs not in tbses:
                rp = pc.get_regulation_periods(lpp.tbs.id, start_date=self.today())

                options.append({
                    'tbs_name': lpp.tbs.name,
                    'tbs_id': lpp.tbs.id,
                    'rp_id': rp.id
                })
                tbses.append(lpp.tbs)
        return options

    def __query_for_companies(self, _where, limit, country_id=None):
        companies = []
        if country_id is not None:
            companies = self.sa_session.query(WeCompany).outerjoin(WeHqppLicense, WeHqppLicense.company_id == WeCompany.id) \
                .outerjoin(WeTbsLicense, WeTbsLicense.company_id == WeCompany.id) \
                .outerjoin(WeLpp, WeLpp.company_id == WeCompany.id) \
                .join(CompanyAddress) \
                .join(Address, and_(CompanyAddress.address_id == Address.id,
                                    Address.country_id == country_id)) \
                .filter(_where).limit(limit).all()
        else:
            companies = self.sa_session.query(WeCompany).outerjoin(WeHqppLicense, WeHqppLicense.company_id == WeCompany.id) \
                .outerjoin(WeTbsLicense, WeTbsLicense.company_id == WeCompany.id) \
                .outerjoin(WeLpp, WeLpp.company_id == WeCompany.id) \
                .filter(_where).limit(limit).all()

        return companies