# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy.sql.expression import and_

from basesoft.web.contrib.contact.controllers import SimpleContactController
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.contrib.order.controllers import AbstractOrderController
from basesoft.web.contrib.order.models import OrderLineRule
from basesoft.web.core.util.controllers import transaction_decorator

from we.core.controllers import AbstractWeController
from we.declaration.models import WeDeclaration, WeDeclarationResubmitNotification
from we.member.models import WeLpp
from we.order.models import WeDeclarationOrder, WeDeclarationOrderLine
from we.product.models import WeHqppRegulationPeriodProduct, WeTbsProductLevel
from we.tbs.models import WeTbsLicense


class WeDeclarationOrderController(AbstractWeController, AbstractOrderController):
    _instance_class = WeDeclarationOrder

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['lpp', 'year', 'month', 'quarter', 'bi_annual']
        }, {
            WeDeclarationOrderLine: ['tbs_product_level', 'weight', 'is_b2b'],
            OrderLineRule: ['tax_class_code', 'gl_account_code'],  # FIXME this should be added through parent!
        })


    def load_state(self, _id, submit_dry_run=False):
        if isinstance(_id, self._instance_class):
            order = _id
        else:
            order = self.get(_id).pop()

        scc = SimpleContactController(self.sa_session, account_id=self.get_account().id)

        # The "client" of the order is the LPP invoice company if it is set, else the LPP itself.
        ic = order.lpp.company if order.lpp.invoice_company is None else order.lpp.invoice_company

        rv = super(WeDeclarationOrderController, self).load_state(order, submit_dry_run)
        rv['year'] = order.year
        rv['month'] = order.month
        rv['quarter'] = order.quarter
        rv['bi_annual'] = order.bi_annual
        rv['client'] = scc.load_company_state(ic)
        rv['has_end_user_distinction'] = order.merchant.get_config('we.tbs.require.end.user.distinction')

        return rv

    def _load_state_for_export(self, order):
        rv = super(WeDeclarationOrderController, self)._load_state_for_export(order)

        lpp_relation_id = order.lpp.lpp_relation_id
        if lpp_relation_id is None or lpp_relation_id.strip() == '':
            lpp_relation_id = None

        invoice_relation_id = order.lpp.invoice_relation_id
        if invoice_relation_id is None or invoice_relation_id.strip() == '':
            invoice_relation_id = None

        rv['properties'] = [{
            'key': 'year',
            'value_int': order.year
        },{
            'key': 'month',
            'value_int': order.month
        },{
            'key': 'quarter',
            'value_int': order.quarter
        },{
            'key': 'bi_annual',
            'value_int': order.bi_annual
        },{
            'key': 'lpp_relation_id',
            'value_str': lpp_relation_id  # this is the ID the TBS should have in their system
        },{
            'key': 'invoice_relation_id',
            'value_str': invoice_relation_id
        }]

        return rv

    def _load_lines_state(self, lines, for_export=False):

        if for_export:
            return self.__load_lines_state_for_export(lines)
        rv = []
        for line in lines:
            rv.append({
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'weight': line.weight,
                'is_b2b': line.is_b2b,
                'tbs_product_level': line.tbs_product_level.label,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
            })
        return rv

    def __load_lines_state_for_export(self, lines):
        rv = []

        for line in lines:

            state_line = {
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': []
            }

            if line.polymorphic_type == 'WeDeclarationOrderLine':

                state_line['properties'] += [{
                        'key': 'weight',
                        'value_decimal': line.weight,
                    },{
                        'key': 'local_category',
                        'value_str': line.tbs_product_level.label,
                    },{
                        'key': 'local_category_description',
                        'value_str': line.tbs_product_level.description,
                    }]

                if line.is_b2b is not WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
                    state_line['properties'].append({
                        'key': 'dimension',
                        'value_str': 'B2B' if line.is_b2b == WeHqppRegulationPeriodProduct.IS_B2B_TRUE else 'B2C'
                    })

            rv.append(state_line)

        return rv

    def _get_line_unit(self, quantity, order, line_polymorphic_type, **kwargs):

        if kwargs.has_key('line'):
            l = kwargs['line']
            tbs_product = l.tbs_product_level
            weight = l.weight
        else:
            tbs_product = kwargs['tbs_product_level']
            weight = kwargs.get('weight')

        if tbs_product.invoice_basis == WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT:
            assert weight is not None
            return weight

        #TODO: Add fixed annual fee

        return int(quantity)  # always force quantity to be an int

    @transaction_decorator
    def process_declarations(self, tbs_id):
        tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.license_id == tbs_id).one()

        # check the day
        closing_day = tbs.get_config('we.tbs.declaration.closing.day')

        values = self.get_tbs_period_for_day(tbs, closing_day)

        if values is None:
            self.logger.info('Today (%s) is not the closing date for TBS %s' % (tbs.domain.today(), tbs))
            return None

        self.logger.info('Declaration closing date reached for TBS %s' % tbs)

        year = values[0]
        biannual = values[1]
        quarter = values[2]
        month = values[3]
        # get all active lpps at this tbs
        lpps = self.sa_session.query(WeLpp).filter(
            and_(
                WeLpp.tbs == tbs,
                WeLpp.status == WeLpp.STATUS_ACTIVE
            )
        )

        rv = []

        for lpp in lpps:
            o = self._process_declarations_for_lpp(lpp, year, biannual, quarter, month)
            if o is not None:
                rv.append(o)
        return rv

    def _process_declarations_for_lpp(self, lpp, year, biannual, quarter, month):
        self.logger.info(' - Processing complete Declarations for LPP %s' % lpp)

        months = self.get_months_for_period(month, quarter, biannual)

        # get all unprocessed declarations per lpp,
        declarations = self.sa_session.query(WeDeclaration).filter(
            and_(
                WeDeclaration.lpp == lpp,
                WeDeclaration.status == WeDeclaration.STATUS_COMPLETE,
                WeDeclaration.declaration_order == None,
                WeDeclaration.year == year,
                WeDeclaration.month.in_(months)
            )
        ).order_by(WeDeclaration.month).all()

        if len(declarations) == 0:
            self.logger.info(' - No completed unexported Declarations found.')
            return None

        self.logger.info(' - %d complete Declarations found. Squashing them per Local Category.' % len(declarations))

        order_lines, quantities = self._aggregate_order_lines_per_local_category(declarations)

        # save order
        order = self._save_or_update_order(None, order_lines, merchant=lpp.tbs, scope=lpp.hqpp.default_group, lpp=lpp,
                                           year=year, bi_annual=biannual, quarter=quarter, month=month, determine_order_type=True)
        self.logger.info(' - Added new order %s' % order)

        for declaration in declarations:
            declaration.status = WeDeclaration.STATUS_EXPORTED
            declaration.declaration_order = order
            self.sa_session.add(declaration)

        return order

    @transaction_decorator
    def process_resubmits(self, tbs_id):
        tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.license_id == tbs_id).one()

        # get all active lpps at this tbs
        lpps = self.sa_session.query(WeLpp).filter(
            and_(
                WeLpp.tbs == tbs,
                WeLpp.status == WeLpp.STATUS_ACTIVE
            )
        )

        rv = []

        for lpp in lpps:
            orders = self._process_resubmits_for_lpp(lpp)
            if orders is None:
                continue
            rv.extend(orders)

        self.create_declaration_resubmit_notification(tbs)
        return rv

    def _process_resubmits_for_lpp(self, lpp):
        """
         Get all declarations for a LPP which are not exported and for each declaration
         get the TBS period it is for, check for an order for that period. If there is an order
         for that period get all declarations for that period and resubmit the order.

        :param lpp: WeLpp object
        :return: List of WeDeclarationOrder or None
        """
        self.logger.info(' - Processing Resubmit for LPP %s' % lpp)

        # get all declarations per lpp which are not exported yet
        declarations = self.sa_session.query(WeDeclaration).filter(
            and_(
                WeDeclaration.lpp == lpp,
                WeDeclaration.status.in_([WeDeclaration.STATUS_COMPLETE, WeDeclaration.STATUS_RESUBMIT]),
            )
        ).order_by(WeDeclaration.month).all()

        if len(declarations) == 0:
            self.logger.info(' - No updated, exported Declarations found.')
            return None

        self.logger.info(' - %d Resubmits found. Squashing them per Local Category.' % len(declarations))

        orders = []
        used_orders = []

        for d in declarations:
            order = d.declaration_order

            values = self.get_tbs_period_for_month(lpp.tbs, d.month, d.year)
            year = values[0]
            biannual = values[1]
            quarter = values[2]
            month = values[3]

            months = self.get_months_for_period(month, quarter, biannual)

            # if declaration has no order but is marked as resubmit we should have an order for that tbs period
            if order is None and d.status == WeDeclaration.STATUS_RESUBMIT:
                order = self.sa_session.query(WeDeclarationOrder).filter(
                    and_(
                        WeDeclarationOrder.lpp == lpp,
                        WeDeclarationOrder.year == year,
                        WeDeclarationOrder.bi_annual == biannual,
                        WeDeclarationOrder.quarter == quarter,
                        WeDeclarationOrder.month == month,
                        WeDeclarationOrder.type != WeDeclarationOrder.TYPE_CREDIT,
                        WeDeclarationOrder.status == WeDeclarationOrder.STATUS_SUBMITTED
                    )
                ).order_by(WeDeclarationOrder.id.desc()).first()

            # if order is still None or order has been processed already
            if order is None or order.id in used_orders:
                continue

            used_orders.append(order.id)

            # get all declarations for the tbs period and make a resubmit
            declars = self.sa_session.query(WeDeclaration).filter(
                and_(
                    WeDeclaration.lpp == lpp,
                    WeDeclaration.year == year,
                    WeDeclaration.month.in_(months)
                )
            ).order_by(WeDeclaration.month).all()

            order_lines, quantities = self._aggregate_order_lines_per_local_category(declars)

            new_order = self.resubmit_order(order, order_lines, quantities, year, biannual, quarter, month)

            orders.append(new_order)

            for dec in declars:
                dec.status = WeDeclaration.STATUS_EXPORTED
                dec.declaration_order = new_order
                self.sa_session.add(dec)
            used_orders.append(new_order.id)

        return None if len(orders) == 0 else orders

    def _aggregate_order_lines_per_local_category(self, declarations):
        lines = {}
        ordered = []
        quantities = {}
        for declaration in declarations:
            for line in declaration.lines:
                rpp = line.hqpp_period_product.tbs_regulation_period_product
                key = (rpp.id, line.hqpp_period_product.is_b2b)
                if key not in ordered:
                    ordered.append(key)
                if not lines.has_key(key):
                    lines[key] = {'rpp': rpp, 'lines': [], 'is_b2b': line.hqpp_period_product.is_b2b}
                lines[key]['lines'].append(line)

        order_lines = []
        for key in ordered:
            product = lines[key]
            rpp = product['rpp']
            ol = {
                'polymorphic_type': WeDeclarationOrderLine,
                'description': '%s - %s' % (rpp.tbs_product_level.label, rpp.tbs_product_level.description),
                'tbs_product_level': rpp.tbs_product_level,
                'price': rpp.fee,
                'quantity': 0,
                'weight': 0,
                'is_b2b': product['is_b2b']
            }
            # sum quantity and weight for all lines with same local category
            for line in product['lines']:
                ol['quantity'] += line.quantity
                ol['weight'] += line.product_total_weight
            quantities['line_quantity_%s' % len(order_lines)] = ol['quantity']
            quantities['line_weight_%s' % len(order_lines)] = ol['weight']
            quantities['line_%s' % len(order_lines)] = rpp.tbs_product_level

            order_lines.append(ol)

        return order_lines, quantities

    @transaction_decorator
    def resubmit_order(self, order, order_lines, quantities, year, biannual, quarter, month):
        """
         Check whether there are different quantities. If there are differences make a credit
         and a new order a child of the credit order.

        :param order: WeDeclarationOrder object
        :param order_lines: Dict
        :param quantities: int
        :param year: int
        :param biannual: int
        :param quarter: int
        :param month: int
        :return: WeDeclarationOrder object
        """
        assert isinstance(order, self._instance_class)

        self.logger.info('WeDeclarationOrderController: resubmit_order() order=%s' % (order.id))

        if not self.has_different_quantities(order, quantities) and order_lines is None:
            self.logger.info(' - order is not changed')
            # make sure the order is set to compliant
            order.is_compliant = WeDeclarationOrder.IS_COMPLIANT
            self.sa_session.add(order)
            return order

        # if order is exported need to resubmit else just update
        if order.export_date != None:
            inverse = self._inverse_of_parent_order(order)
            order = self._create_child_order(inverse, include_line_types=[WeDeclarationOrderLine])

        lpp = order.lpp

        return self._save_or_update_order(order.id, order_lines, merchant=lpp.tbs, scope=lpp.hqpp.default_group,
                                          lpp=lpp, year=year, bi_annual=biannual, quarter=quarter, month=month, determine_order_type=True)

    def has_different_quantities(self, order, values):
        """
        Overrides super and also checks for weight/etc
        """
        if len(order.lines) != len(values.values()): # different sizes means different quantities.
            return True

        # if any of the user fields (quantity, weight) changes, make a credit regardless which field
        # is used for calculation
        for line in order.lines:
            # check whether local category is different
            if values['line_%s' % line.index] != line.hqpp_period_product.tbs_regulation_period_product.tbs_product_level:
                return True

            for field in line.CREDIT_FIELDS:
                if hasattr(line, field):
                    v = values['line_%s_%s' % (field, line.index)]
                    if getattr(line, field) != v:
                        return True

        return False

    def create_declaration_resubmit_notification(self, tbs_obj):
        nc = NotificationController(self.sa_session)
        callback = {
            'js_module': 'we_declaration_manager'
        }
        notification = nc.set_notification('', '', instance_class=WeDeclarationResubmitNotification,
                                           callback=callback)
        nc.create_system_notification(notification, tbs_obj.default_group)
