'''
Created on Jan 14, 2015

@author: rado
'''

import datetime
import os

from decimal import Decimal, InvalidOperation
from openpyxl.cell import get_column_letter
from django.utils.translation import ugettext as _, ugettext_lazy
from sqlalchemy.sql.expression import and_

from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core import L10n_settings as l10n
from basesoft.web.contrib.file.controllers import ImportFileController, ImportFileException
from basesoft.web.core.importers import Utf8CsvImporter, XlsxImporter

from iqportal.weee.core.models import WEEEProductLevel

from we.member.models import WeHqppLicense, WeLpp
from we.product.controllers import WeProductController
from we.product.models import WeHqppProduct, WeTbsProductLevel, WeUnuCode, WeCN8, \
    WeCN8PeriodProduct, WeTbsRegulationPeriodProduct, WeHqppRegulationPeriodProduct
from we.tbs.models import WeTbsLicense, WeTbsRegulationPeriod


class WeImportException(ImportFileException):
    pass


class CN8AbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.product.CN8ToWeeeCategory'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('CN8 - WEEE - UNU')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = CN8CsvImporter

        elif file_extension == '.xlsx':
            importer = CN8XlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

class CN8LineImporter(CN8AbstractImporter):
    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = 5
    weee_cat_map = None
    unu_code_map = None

    def __init__(self, session, **kwargs):
        ImportFileController.__init__(self, session, **kwargs)
        self.reset()

    def reset(self):
        CN8AbstractImporter.reset(self)
        self.weee_cat_map = {}
        self.unu_code_map = {}

    def _import_row(self, row, controller, dryrun=False, **kwargs):
        try:
            int(row[0])
        except ValueError:
            raise WeImportException(_('Invalid CN8 code (%s) found, must be numerical.' % row[0]))

        cn8 = self.sa_session.query(WeCN8).filter(WeCN8.code == int(row[0])).first()

        if row[2] not in [None, '']:
            # weee category label
            row_weee_cat = row[2]
            if not self.weee_cat_map.has_key(row_weee_cat):
                weee_cat = self.sa_session.query(WEEEProductLevel).filter(WEEEProductLevel.label == row_weee_cat).first()
                if weee_cat is None:
                    raise WeImportException(_('WEEE category with label %s does not exist.' % row_weee_cat))
                self.weee_cat_map[weee_cat.label] = weee_cat.id

            # unu code
            row_unu = row[3].strip()
            row_unu_desc = row[4].strip()
            if row_unu == '':
                self.unu_code_map[row_unu] = None
            elif not self.unu_code_map.has_key(row_unu):
                unu_code = self.sa_session.query(WeUnuCode).filter(WeUnuCode.code == row_unu).first()
                if unu_code is None:
                    unu_code = WeUnuCode()
                    unu_code.code = row_unu
                    unu_code.description = row_unu_desc
                    self.sa_session.add(unu_code)
                    self.sa_session.flush()
                self.unu_code_map[unu_code.code] = unu_code.id

            if cn8 is None:

                sql = """
                      INSERT INTO tblWeCN8(code, unu_code_id, weee_product_level_id, description, created, last_updated) VALUES
                      (:cn8_code, :unu_code_id, :weee_product_id, :description, NOW(), NOW())
                      """
            else:
                sql = """
                        UPDATE tblWeCN8 SET unu_code_id = :unu_code_id, weee_product_level_id = :weee_product_id,
                        description = :description, last_updated= NOW() WHERE code = :cn8_code;
                        """
            values = {
                'unu_code_id': self.unu_code_map[row_unu],
                'weee_product_id': self.weee_cat_map[row_weee_cat],
                'description': row[1],
                'cn8_code': int(row[0])
            }
            self.sa_session.execute(sql, values)
        else:
            # weee category and unu are empty
            if cn8 is None:
                sql = """
                      INSERT INTO tblWeCN8(code, description, created, last_updated) VALUES(:cn8_code, :description, NOW(), NOW())
                      """
            else:
                sql = """
                        UPDATE tblWeCN8 SET description = :description, last_updated = NOW() WHERE code = :cn8_code;
                        """

            values = {
                'description': row[1],
                'cn8_code': int(row[0])
            }
            self.sa_session.execute(sql, values)
        return cn8

    @staticmethod
    def get_notification_text():
        return ''

    @staticmethod
    def get_notification_subject():
        return 'CN8 to WEEE category import'

    @staticmethod
    def get_failure_notification_callback():
        return None

    @staticmethod
    def get_notification_callback():
        return {
            'js_module': 'we_cn8_weee_product_level_manager'
        }

    @staticmethod
    def get_notification_update_notify():
        return ['we.product.FileImportCustomQuery.newData', 'we.product.WeCN8WeeeProductLevelCustomQuery.newData']


class CN8CsvImporter(CN8LineImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        for row in self._read_row(file_path):
            yield self.get_only_expected_columns(row)


class CN8XlsxImporter(CN8LineImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):

        active_sheet = self._read_row(file_path)

        for row in active_sheet.iter_rows():
            values = []

            row = self.get_only_expected_columns(row)
            for cell in row:

                value = cell.internal_value
                if value is None:
                    value = ''
                if type(value) == float:
                    value = int(value)
                values.append(value)

            yield values


class HqppProductAbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.product.HqppProduct'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('Product sheet')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = HqppProductCsvImporter

        elif file_extension == '.xlsx':
            importer = HqppProductXlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_human_readable_settings(self, import_settings):
        rv = {}

        pc = WeProductController(self.sa_session, account_id=self.get_account().id)
        hqpp = pc.get_member_licenses(import_settings['hqpp_id']).pop()

        rv[_('HQPP')] = hqpp.name
        rps = []
        for tbs, rp_id in import_settings['regulation_periods'].iteritems():
            tbs = pc.get_tbs(tbs[4:])
            rp = pc.get_regulation_period_by_id(tbs.id, rp_id).title
            rps.append('%s: %s' % (tbs.name, rp))

        rv[_('Regulation periods')] = '; '.join(rps)
        return rv


class HqppProductLineImporter(HqppProductAbstractImporter):

    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = None

    rp_dict = None
    tbs_dict = None
    callback = None

    def reset(self):
        HqppProductAbstractImporter.reset(self)
        self.rp_dict = {}
        self.tbs_dict = {}
        self.first_row = True
        self.headers = 0

    def set_notification_callback(self):
        self.callback = {
            'js_module': 'we_hqpp_product_manager'
        }

    def get_notification_callback(self):
        return self.callback

    @staticmethod
    def get_failure_notification_callback():
        return {
            'js_module': 'we_hqpp_product_sheet_import'
        }

    @staticmethod
    def get_notification_subject():
        return None

    @staticmethod
    def get_notification_text():
        return None

    @staticmethod
    def get_notification_update_notify():
        return ['we.product.FileImportCustomQuery.newData', 'we.product.WeHqppProductCustomQuery.newData']

    def _validate_row_headers(self):
        required = ['sku', 'name', 'description', 'weight_per_unit', 'length_per_unit', 'height_per_unit', 'width_per_unit',
            'diagonal', 'weight_cat1', 'weight_cat2', 'weight_cat3', 'weight_cat4', 'weight_documentation', 'weight_batteries',
            'weight_package', 'cn8', 'is_b2b']
        not_found = []
        for i in required:
            if i in self._file_headers:
                continue
            not_found.append(i)
        if len(not_found) > 0:
            raise WeImportException(_('The following columns were not found: %(columns)s' % {'columns': ', '.join(not_found)}))

    def _import_row(self, row, controller, dryrun, hqpp_id=None, regulation_periods=None):
        if self.first_row is True:
            self.first_row = False
            self.set_notification_callback()

        self._validate_row_headers()

        hqpp = self.sa_session.query(WeHqppLicense).filter(WeHqppLicense.id==hqpp_id).one()
        product = self.sa_session.query(WeHqppProduct)\
            .filter(and_(WeHqppProduct.sku == row[0],
                         WeHqppProduct.hqpp == hqpp)).first()
        values = {}
        tbs_rpp = []
        is_b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE

        exceptions = []
        for i in xrange(len(self._file_headers)):
            # if column has no value skip it
            if row[i] in [None, '']:
                continue

            # columns from 4 to 15 are weight columns
            if 3 < i < 16:
                try:
                    row[i] = Decimal(row[i])
                except InvalidOperation:
                    raise WeImportException('Value %s cannot be converted to a decimal number. '
                                            'The decimal separator should be a dot and a thousand '
                                            'separator should not be used.' % row[i])

            h = self._file_headers[i].lower()

            if h == 'cn8':
                cn8 = self.sa_session.query(WeCN8).filter(WeCN8.code == row[i]).first()
                if cn8 is None:
                    raise WeImportException(_("CN8 code %s does not exist in the database. "
                                              "Please add it before importing again.") % row[i])

            if h == 'is_b2b':
                is_b2b_val = row[i]
                if is_b2b_val.lower() == 'b2b':
                    is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_TRUE
                elif is_b2b_val.lower() == 'b2c':
                    is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_FALSE
                else:
                    is_b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE
                continue

            # if headers starts with tbs- get the number and then tbs object
            if len(h) > 4 and h[:4] == 'tbs-':

                # We store the exceptions on a per TBS basis to provide proper feedback to the user.

                tbs_id = int(h[4:])
                tbs_product_label = row[i]

                if not self.tbs_dict.has_key(h):
                    tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.license_id == tbs_id).first()
                    if tbs is None:
                        exceptions.append(_("There is no such TBS with ID %s ") % tbs_id)
                        continue
                    self.tbs_dict[h] = tbs
                else:
                    tbs = self.tbs_dict[h]

                    # check that HQPP has LPPs for that TBS
                    lpps = []
                    for lpp in hqpp.lpps:
                        if lpp.status == WeLpp.STATUS_ACTIVE and lpp.tbs.id == tbs.id:
                            lpps.append(lpp)
                    if len(lpps) == 0:
                        exceptions.append(_("Producer %(producer)s has no local producers registered "
                                                  "at take-back scheme %(tbs)s." % {'producer': hqpp.name, 'tbs': tbs.name}))
                        continue

                distinction = tbs.get_config('we.tbs.require.end.user.distinction')
                if distinction and is_b2b is WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
                    exceptions.append(_('Take-back scheme "%s" requires end-user distinction (B2B/B2C) but row is not indicated as either.' % tbs.name))
                    continue
                elif not distinction and is_b2b is not WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
                    exceptions.append(_('Take-back scheme "%s" does not require end-user distinction (B2B/B2C) but row is indicated as such.' % tbs.name))
                    continue

                # check whether the user has selected a regulation period for tbs
                if not regulation_periods.has_key(h):
                    exceptions.append(_("A regulation period for take-back scheme %s is not selected. "
                                        "Please select one and import again." % tbs.name))
                    continue

                # check whether the regulation period exists
                if not self.rp_dict.has_key(h):
                    rp_id = regulation_periods[h]
                    rp = self.sa_session.query(WeTbsRegulationPeriod)\
                        .filter(and_(WeTbsRegulationPeriod.id == rp_id,
                                     WeTbsRegulationPeriod.tbs_id == tbs.id,
                                     WeTbsRegulationPeriod.archived == 0)).first()
                    if rp is None:
                        exceptions.append(_("Regulation period with ID %(id)s for take-back scheme %(tbs)s "
                                            "does not exist. Either select another one or create one." % {'id': rp_id, 'tbs': tbs.name}))
                        continue
                    self.rp_dict[h] = rp
                else:
                    rp = self.rp_dict[h]

                # check whether local category exists
                pl = self.sa_session.query(WeTbsProductLevel)\
                    .filter(WeTbsProductLevel.tbs_id == tbs_id,
                            WeTbsProductLevel.label == tbs_product_label).first()
                if pl is None:
                    exceptions.append(_('There is no such category with '
                                              'label "%(label)s" at TBS with ID %(id)s ') % {'label': row[i], 'id': tbs_id})
                    continue

                # check whether local category is part of regulation period
                trpp = self.sa_session.query(WeTbsRegulationPeriodProduct).join(WeTbsProductLevel)\
                    .filter(WeTbsProductLevel.tbs_id == tbs_id,
                            WeTbsProductLevel.label == tbs_product_label,
                            WeTbsRegulationPeriodProduct.tbs_regulation_period == rp).first()
                if trpp is None:
                    exceptions.append(_("TBS local category %(category)s is not part of regulation period %(period)s. "
                                              "Please make sure all local categories are already "
                                              "in the system and are part of the regulation period "
                                              "before importing.") % {'category': row[i], 'period': rp.label})
                    continue
                tbs_rpp.append(trpp)
                continue

            values[h] = row[i]
            values['created'] = datetime.datetime.now()
            values['last_updated'] = datetime.datetime.now()
            values['hqpp_id'] = hqpp.id

        if len(exceptions) > 0:
            raise WeImportException(' | '.join(exceptions))

        if product is None:
            args = values.keys()
            vals = [':' + val for val in args]

            args = '(' + ','.join(args) + ')'
            vals = '(' + ','.join(vals) + ')'

            sql = """
                  INSERT INTO tblWeHqppProduct %s
                  VALUES %s
                  """ % (args, vals)
            self.sa_session.execute(sql, values)
            # get the newly saved product and link it to TBS product levels
            product = self.sa_session.query(WeHqppProduct)\
                .filter(and_(WeHqppProduct.sku == values['sku'],
                             WeHqppProduct.hqpp == hqpp)).first()

            for l in tbs_rpp:
                hrpp = WeHqppRegulationPeriodProduct()
                hrpp.product = product
                hrpp.tbs_regulation_period_product = l
                hrpp.tbs_regulation_period = l.tbs_regulation_period
                if is_b2b is not None:
                    hrpp.is_b2b = is_b2b
                self.sa_session.add(hrpp)
            self.sa_session.flush()
            return product

        args = values.keys()
        params = ''
        for k in args:
            if k == 'sku' or k == 'hqpp_id':
                continue
            params += '%s = :%s,' % (k, k)
        # remove the comma at the end
        params = params[:len(params) - 1]

        sql = """
                UPDATE tblWeHqppProduct SET %s WHERE sku = :sku AND hqpp_id = :hqpp_id;
                """ % params
        self.sa_session.execute(sql, values)

        for trpp in tbs_rpp:
            rp = trpp.tbs_regulation_period
            # get the link of tbs regulation product to hqpp product for the given regulation product
            hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct)\
                .filter(and_(WeHqppRegulationPeriodProduct.product_id == product.id,
                             WeHqppRegulationPeriodProduct.is_b2b == is_b2b,
                             WeHqppRegulationPeriodProduct.tbs_regulation_period == rp)).first()

            if hrpp is None:
                # add new one
                hrpp = WeHqppRegulationPeriodProduct()
                hrpp.product = product
                hrpp.tbs_regulation_period_product = trpp
                hrpp.tbs_regulation_period = trpp.tbs_regulation_period
                if is_b2b is not None:
                    hrpp.is_b2b = is_b2b
            else:
                # update
                hrpp.tbs_regulation_period_product = trpp
            self.sa_session.add(hrpp)
        self.sa_session.flush()
        return product


class HqppProductCsvImporter(HqppProductLineImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        for row in self._read_row(file_path):
            self.headers += 1
            if self.headers == 2:
                self.NUMBER_EXPECTED_COLUMNS = len(row)
                continue
            yield self.get_only_expected_columns(row)

class HqppProductXlsxImporter(HqppProductLineImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):

        active_sheet = self._read_row(file_path)
        row_i = 1
        for row in active_sheet.iter_rows():
            values = []
            cell_i = 1
            if row_i == 1:
                self.NUMBER_EXPECTED_COLUMNS = len(row)

            if row_i == 2:
                row_i += 1
                continue

            row_i += 1

            row = self.get_only_expected_columns(row)
            for cell in row:
                value = cell.internal_value
                if value is None:
                    value = ''
                # make cn8 an int not float
                if cell_i == 16 and type(value) == float:
                    value = int(value)
                if cell_i > 16 and value != '':
                    if not isinstance(value, unicode):
                        WeImportException("Column '%s' "
                                          "must be formatted to text." % get_column_letter(cell_i))
                values.append(value)
                cell_i += 1

            yield values


class TbsPriceSheetAbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.product.TbsPriceSheet'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('TBS price sheet')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = TbsPriceSheetCsvImporter

        elif file_extension == '.xlsx':
            importer = TbsPriceSheetXlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_human_readable_settings(self, import_settings):
        rv = {}

        pc = WeProductController(self.sa_session, account_id=self.get_account().id)
        tbs = pc.get_tbs(import_settings['tbs_id'])

        # FIXME, weight settings?
        rv[_('Take-back scheme')] = tbs.name
        if import_settings['regulation_period_id'] is not None:
            rp = pc.get_regulation_period_by_id(tbs.id, import_settings['regulation_period_id'])
            label = rp.label
            year = rp.start_date.year
            month = rp.start_date.month
        else:
            label = import_settings['regulation_label']
            year = import_settings['regulation_year']
            month = import_settings['regulation_month']

        rv[_('Regulation period')] = label
        rv[_('Start year')] = year
        rv[_('Start month')] = l10n.default.MONTHS_LONG[int(month)-1]
        return rv

class TbsPriceSheetLineImporter(TbsPriceSheetAbstractImporter):

    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = 5
    number_of_rows = 1
    local_categories = None
    new_categories = None
    callback = None
    _subject = ''
    _notification_text = ''

    def reset(self):
        TbsPriceSheetAbstractImporter.reset(self)
        self.number_of_rows = 1  # reset the number of rows.
        self.local_categories = []
        self.new_categories = []
        self._subject = 'New regulation period'
        self._notification_text = 'Use the perform action button to view the changes compared to the previous period'
        self._parents = {}

    def set_notification_callback(self, rp_id, is_new=False):
        if not is_new:
            self.callback = {
                'js_module': 'we_tbs_regulation_period_compare',
                'parameters': [rp_id]
            }
            return
        # if this is a new Regulation Period, we open the Suggestions upload.
        self.callback = {
            'js_module': 'we_cn8_tbs_product_level_import',
            'parameters': []
        }

    def get_notification_callback(self):
        return self.callback

    @staticmethod
    def get_failure_notification_callback():
        return {
            'js_module': 'we_tbs_price_sheet_import'
        }

    def get_notification_subject(self):
        return self._subject

    def get_notification_text(self):
        return self._notification_text

    def set_notification_texts(self, rp, is_new=False):
        if is_new:
            subject = 'New regulation period "%(rp)s" for TBS "%(tbs)s"'
            text = 'A new regulation period "%(rp)s" has been added for Take-back scheme "%(tbs)s". Use the perform action button to upload the CN8 based suggestions.'
        else:
            subject = 'Updated regulation period "%(rp)s" for TBS "%(tbs)s"'
            text = 'The regulation period "%(rp)s" at Take-back scheme "%(tbs)s" has been updated. Use the perform action button to view the changes compared to the previous period.'

        self._notification_text = text % {'rp': rp.label, 'tbs': rp.tbs.name}
        self._subject = subject % {'rp': rp.label, 'tbs': rp.tbs.name}

    @staticmethod
    def get_notification_update_notify():
        return ['we.product.FileImportCustomQuery.newData', 'we.product.WeTbsPriceSheetCustomQuery.newData']

    def _import_row(self, row, controller, dryrun, tbs_id=None, regulation_period_id=None,
                    regulation_label=None, regulation_year=None, regulation_month=None, weight_settings=None,
                    row_count=None, **kwargs):
        parent_level = None
        parent_level_id = None
        main_category = None if row[0] in [None, ''] else unicode(row[0]).strip()
        category_label = unicode(row[1]).strip()
        description = unicode(row[2]).strip()
        invoice_basis = row[3]
        fee = row[4]
        if type(fee) in (str, unicode, ):
            fee = fee.strip()
            if fee == '':
                fee = None
        five_places = Decimal(10) ** -5
        self.number_of_rows += 1

        if category_label == main_category:
            raise WeImportException(_('Subcategory %(sub)s cannot have the same label as its'
                                      ' main category %(main)s.' % {'sub': category_label, 'main': main_category}))

        tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.id == tbs_id).first()

        if tbs is None:
            raise WeImportException(_('There is no TBS in the system with ID %s' % tbs_id))

        if main_category is not None:
            # Get the main category from the locally stored list of main categories.
            if not self._parents.has_key(main_category):
                raise WeImportException(_("The main category was not found in the file preceding this record. "
                                          "Make sure the order of the categories is correct."))
            parent_level = self._parents[main_category]
            parent_level_id = parent_level.id

        product_level = self.sa_session.query(WeTbsProductLevel)\
            .filter(and_(WeTbsProductLevel.tbs == tbs,
                         WeTbsProductLevel.label == category_label,
                         WeTbsProductLevel.parent_level == parent_level)).first()

        if product_level is not None and product_level.description.lower() != description.lower():
            product_level = None
        elif product_level is not None and product_level.archived:  # previously archived product levels should be unarchived in case they are reused.
            controller.logger.info('Product level (%s) unarchived by system due to new import.' % product_level)
            product_level.archived = False
            self.sa_session.add(product_level)

        if fee is not None:
            fee = str(fee)
            try:
                fee = Decimal(fee).quantize(five_places)
            except InvalidOperation:
                raise WeImportException('The fee provided is not a valid decimal number, '
                                        'please remove any currency symbols.')
            if fee.is_signed() and not tbs.get_config('we.tbs.use.negative.fee'):
                raise WeImportException(_("TBS %s does not allow negative fees." % tbs.name))

        if invoice_basis in [None, '']:
            # if parent_level is None current row is a main category
            if parent_level is None:
                raise WeImportException('Missing basis for invoice. Parent does not have one either.')
            invoice_basis = parent_level.invoice_basis
        else:
            if invoice_basis.lower() == 'quantity':
                invoice_basis = WeTbsProductLevel.INVOICE_BASED_ON_QUANTITY
            elif invoice_basis.lower() == 'weight':
                invoice_basis = WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT
            elif invoice_basis.lower() == 'fixed annual fee':
                invoice_basis = WeTbsProductLevel.INVOICE_BASED_ON_FIXED_ANNUAL_FEE
            else:
                raise WeImportException(_("Wrong value '%s' for basis for invoice. "
                                          "Use 'quantity', 'weight' or 'fixed annual fee'." % invoice_basis))

        if fee is None and main_category is not None:
            f = None
            # get the fee from the parent category

            for i in xrange(1, len(self.local_categories) + 1):
                if self.local_categories[-i]['id'] == parent_level.id:
                    f = self.local_categories[-i]['fee']
                    break
            # if parent category is not in the import list then ask user for fee
            if f is None:
                raise WeImportException('Fee should be provided for this category '
                                        'because the main category fee is not listed in the price sheet.')
            fee = f

        category_key = (category_label, parent_level_id)
        if category_key in self.new_categories:
            raise WeImportException(_("You cannot have two categories with the same label '%(label)s' "
                                      "and the same parent category ID '%(id)s' "
                                      "in the same regulation period." % {'label': category_label, 'id': parent_level_id}))

        self.new_categories.append(category_key)

        # add new level
        if product_level is None:
            product_level = controller.save_or_update_tbs_category(None, category_label,
                                                                   description, tbs,
                                                                   parent_level_id=parent_level_id,
                                                                   invoice_basis=invoice_basis)
        if parent_level is None:
            self._parents[product_level.label] = product_level

        controller.logger.info('Tbs product level [%s] imported.' % category_label)
        # create dictionary for adding to regulation period
        self.local_categories.append({'id': product_level.id, 'fee': fee, 'row': row, 'line_number': self.number_of_rows})

        return True

    def _post_read_rows(self, controller, tbs_id=None, regulation_period_id=None,
                        regulation_label=None, regulation_year=None, regulation_month=None,
                        weight_settings=None, **kwargs):

        if regulation_period_id is None:
            regulation_year = int(regulation_year)
            regulation_month = int(regulation_month)
            weight_settings = int(weight_settings)
        else:
            rp = controller.get_regulation_period_by_id(int(tbs_id), regulation_period_id)
            regulation_year = rp.start_date.year
            regulation_month = rp.start_date.month
            regulation_label = rp.label
            weight_settings = rp.weight_settings

        try:
            rp = controller \
                .save_or_update_regulation_period(regulation_period_id, tbs_id, regulation_label,
                                                  regulation_year, regulation_month, weight_settings,
                                                  self.local_categories, return_errors=True)
        except PublicUserException as e:
            self._failure_lines += e.args[1]

        controller.logger.info(' - Finished importing tbs product levels '
                               'for regulation period [%s].' % rp.label)
        self.set_notification_callback(rp.id, regulation_period_id is not None)
        self.set_notification_texts(rp, regulation_period_id is not None)


class TbsPriceSheetCsvImporter(TbsPriceSheetLineImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        for row in self._read_row(file_path):
            yield self.get_only_expected_columns(row)


class TbsPriceSheetXlsxImporter(TbsPriceSheetLineImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):

        active_sheet = self._read_row(file_path)

        for row in active_sheet.iter_rows():
            values = []
            cell_i = 0
            row = self.get_only_expected_columns(row)
            for cell in row:
                cell_i += 1

                value = cell.internal_value

                if cell_i == 1 and value is not None:
                    if not isinstance(value, unicode):
                        WeImportException("Column 'Main category label' must be formatted to text.")
                if cell_i == 2:
                    if not isinstance(value, unicode):
                        WeImportException("Column 'Category label' must be formatted to text.")

                values.append(value)

            yield values


class CN8ToTbsProductLevelAbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.product.CN8ToTbsProductLevel'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('CN8 suggestions')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = CN8ToTbsProductLevelCsvImporter

        elif file_extension == '.xlsx':
            importer = CN8ToTbsProductLevelXlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_human_readable_settings(self, import_settings):
        rv = {}

        pc = WeProductController(self.sa_session, account_id=self.get_account().id)
        tbs = pc.get_tbs(import_settings['tbs_id'])

        rv[_('Take-back scheme')] = tbs.name
        rp = pc.get_regulation_period_by_id(tbs.id, import_settings['regulation_period_id'])
        rv[_('Regulation period')] = rp.title
        return rv


class CN8ToTbsProductLevelLineImporter(CN8ToTbsProductLevelAbstractImporter):
    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = 3

    existing_suggestions = None

    def _import_file(self, file_path, controller, dryrun, tbs_id, regulation_period_id, **kwargs):
        # clear all records for the given tbs_id before importing
        controller.logger.info(' - Deleting obsolete lines from suggestion table.')
        if not dryrun:
            self.sa_session.execute("""
                DELETE tblWeCN8PeriodProduct FROM tblWeCN8PeriodProduct
                JOIN tblWeTbsRegulationPeriodProduct
                    ON tblWeCN8PeriodProduct.period_product_id = tblWeTbsRegulationPeriodProduct.id
                JOIN tblWeTbsRegulationPeriod
                    ON tblWeTbsRegulationPeriodProduct.tbs_regulation_period_id = tblWeTbsRegulationPeriod.id
                WHERE tblWeTbsRegulationPeriod.id = :rp_id
                    AND tblWeTbsRegulationPeriod.tbs_id = :tbs_id
            """, {
                'tbs_id': tbs_id,
                'rp_id': regulation_period_id
            })
        return CN8ToTbsProductLevelAbstractImporter._import_file(self, file_path, controller, dryrun, tbs_id=tbs_id, regulation_period_id=regulation_period_id, **kwargs)

    def reset(self):
        CN8ToTbsProductLevelAbstractImporter.reset(self)
        self.existing_suggestions = []

    def _import_row(self, row, controller, dryrun, tbs_id, regulation_period_id):
        cn = int(row[0])
        pl = row[1]
        is_b2b = None
        if len(row) > 2:
            is_b2b = row[2]

        if is_b2b in [None, '']:
            is_b2b = WeCN8PeriodProduct.B2B_NOT_APPLICABLE
        elif is_b2b.lower() == 'b2b':
            is_b2b = WeCN8PeriodProduct.IS_B2B_TRUE
        else:
            is_b2b = WeCN8PeriodProduct.IS_B2B_FALSE

        # get the cn8 else exception
        cn8 = self.sa_session.query(WeCN8).filter(WeCN8.code == cn).first()
        if cn8 is None:
            raise WeImportException(_("CN8 code %s is not in the system. "
                                      "Please add it before importing." % row[0]))

        # check for TBS
        tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.id == tbs_id).first()

        if tbs is None:
            raise WeImportException(_('There is no TBS in the system with ID %s' % tbs_id))

        rp = self.sa_session.query(WeTbsRegulationPeriod).filter(
            and_(
                WeTbsRegulationPeriod.tbs == tbs,
                WeTbsRegulationPeriod.id == regulation_period_id,
                WeTbsRegulationPeriod.archived == 0
            )
        ).first()

        if rp is None:
            raise WeImportException(_('There is no Regulation Period with ID %(id)d for this tbs %(tbs)s' % {'id': regulation_period_id, 'tbs': tbs.name}))

        # get existing local category else exception
        product_level = self.sa_session.query(WeTbsProductLevel).join(WeTbsRegulationPeriodProduct).filter(
            and_(
                WeTbsProductLevel.tbs == tbs,
                WeTbsProductLevel.label == unicode(pl),
                WeTbsRegulationPeriodProduct.tbs_regulation_period == rp
            )
        ).first()
        if product_level is None:
            raise WeImportException(_('TBS local category %s does not exist for this Regulation Period. '
                                      'Please add it before importing.' % pl))

        pp = [p for p in rp.period_products if p.tbs_product_level == product_level].pop()

        r = (cn8.code, product_level.label, is_b2b)
        if r in self.existing_suggestions:
            raise WeImportException(_("There are duplicate entries for CN8 %s. "
                                      "Make sure the CN8 is not linked "
                                      "to the same category twice or that B2B/B2C "
                                      "is different is applicable." % cn8.code))
        self.existing_suggestions.append(r)

        link = WeCN8PeriodProduct()
        link.cn8_code = cn8.code
        link.period_product = pp
        link.is_b2b = is_b2b
        self.sa_session.add(link)

        return cn8

    @staticmethod
    def get_notification_text():
        return ''

    @staticmethod
    def get_notification_subject():
        return 'CN8 to TBS local category import'

    @staticmethod
    def get_failure_notification_callback():
        return None

    @staticmethod
    def get_notification_callback():
        return None

class CN8ToTbsProductLevelCsvImporter(CN8ToTbsProductLevelLineImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        for row in self._read_row(file_path):
            yield self.get_only_expected_columns(row)


class CN8ToTbsProductLevelXlsxImporter(CN8ToTbsProductLevelLineImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):

        active_sheet = self._read_row(file_path)

        for row in active_sheet.iter_rows():
            values = []
            cell_i = 0
            row = self.get_only_expected_columns(row)
            for cell in row:
                cell_i += 1

                value = cell.internal_value

                if cell_i == 1 and value is not None:
                    if not isinstance(value, unicode):
                        WeImportException("Column 'Main category label' must be formatted to text.")
                if cell_i == 2:
                    if not isinstance(value, unicode):
                        WeImportException("Column 'Category label' must be formatted to text.")

                values.append(value)

            yield values


import_file_map = {
    'we.product.CN8ToWeeeCategory': CN8AbstractImporter,
    'we.product.HqppProduct': HqppProductAbstractImporter,
    'we.product.TbsPriceSheet': TbsPriceSheetAbstractImporter,
    'we.product.CN8ToTbsProductLevel': CN8ToTbsProductLevelAbstractImporter,
}