# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.
import datetime
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy import Integer, Column, ForeignKey, Unicode, Date, Boolean
from sqlalchemy.orm import mapper, relation
from sqlalchemy.types import SmallInteger, DateTime

from basesoft.web.contrib.license import models as license_models
from basesoft.sqlalchemy.table import IQTable

from we.tbs import models as tbs_models

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    tblWeTbsLicense = IQTable('tblWeTbsLicense', metadata,
        Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True)
    )

    tblWeTbsRegulationPeriod = IQTable('tblWeTbsRegulationPeriod', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('tbs_id', Integer, ForeignKey('tblWeTbsLicense.license_id', ondelete='CASCADE')),
        Column('label', Unicode(255), nullable=False),
        Column('weight_settings', Integer, nullable=False, default=tbs_models.WeTbsRegulationPeriod.WEIGHT_PER_UNIT),
        Column('start_date', Date, nullable=False, index=True),
        Column('is_processed', Boolean, nullable=False, default=False),

        Column('deleted', SmallInteger, nullable=False, default=0),
        Column('archived', SmallInteger, nullable=False, default=0),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('tbs_id', 'start_date')
    )

    tables['tblWeTbsLicense'] = tblWeTbsLicense
    tables['tblWeTbsRegulationPeriod'] = tblWeTbsRegulationPeriod

    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    mapper(tbs_models.WeTbsLicense, tables['tblWeTbsLicense'],
        polymorphic_identity='WeTbsLicense',
        inherits=license_models.License
    )

    mapper(tbs_models.WeTbsRegulationPeriod, tables['tblWeTbsRegulationPeriod'],
        properties = {
            'tbs': relation(tbs_models.WeTbsLicense),
        }
    )

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'WEEE Europe TBS')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [

        license_models.Permission(u'we_tbs_license', u"""
            This gives the right to create WE TBS licenses
            """, _("WE TBS license management")),


    ]

def app_configs():
    from basesoft.web.contrib.account import models as account_models
    configs = []

    # Can be monthly, yearly, (quarterly)
    config = account_models.Config()
    config.name = 'we.tbs.declaration.period'
    config.type = 'integer'
    config.default = tbs_models.WeTbsLicense.DECLARATION_PERIOD_MONTHLY
    config.description = u'Maximum period in which a declaration has to happen.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'we.tbs.use.sales.price'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Swiss TBSes use sales price to allocate SKU into a local category.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'we.tbs.require.end.user.distinction'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Member needs to provide a TBS category distinction between b2b and b2c.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'we.tbs.use.negative.fee'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Enable negative fee.'
    configs.append(config)

    # the number of days before the Declaration is automatically closed (or forfeited) after it becomes available.
    config = account_models.Config()
    config.name = 'we.tbs.declaration.closing.day'
    config.type = 'integer'
    config.default = 10
    config.description = u'Declaration closing day.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'we.tbs.declaration.reminder.day'
    config.type = 'integer'
    config.default = 9
    config.description = u'Declaration reminder day.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'we.tbs.is.partner'
    config.type = 'boolean'
    config.default = 'True'
    config.description = u'Is this TBS a WEEE-EU partner.'
    configs.append(config)

    return configs

def app_roles():
    return [
        {
            'name': 'WE TBS',
            'required_permission': 'we_manage_roles',
            'permissions': [
                'view_dashboard','save_dashboard',
                'receive_announcements', 'set_custom_query_preset',
                'view_notification', 'archive_notification',
                'we_view_index_widget',
                'we_view_weee_category',
                'we_view_lpp_companies',
                'we_view_tbs_product_level',
                'we_view_declaration',
                'we_view_products',
                'we_view_declaration_order', 'view_order_details',
                'soap_login',
                'view_datasource', 'view_datasource_widget', 'we_view_report',
            ],
            'license_configs': [
                ('basesoft.web.contrib.order.export_zero_amount_orders', 'True'),
            ],
            'account_configs': [
                ('basesoft.account.use.iconbar', 'False'),
            ]
        }
    ]
